r"""Finders
===========
"""

from dataclasses import dataclass

from lsp_tree_sitter.finders import ErrorFinder, QueryFinder, SchemaFinder
from lsprotocol.types import DiagnosticSeverity

from .schema import TmuxTrie
from .utils import get_query, get_schema


@dataclass(init=False)
class ImportTmuxFinder(QueryFinder):
    r"""Import Tmux finder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: found",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ):
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        """
        query = get_query("import")
        super().__init__(query, message, severity)


@dataclass(init=False)
class TmuxFinder(SchemaFinder):
    r"""Tmuxfinder."""

    def __init__(self) -> None:
        r"""Init.

        :rtype: None
        """
        self.validator = self.schema2validator(get_schema())
        self.cls = TmuxTrie


DIAGNOSTICS_FINDER_CLASSES = [
    ErrorFinder,
    TmuxFinder,
]



r"""Misc
========
"""

import re
from typing import Any

from bs4.element import NavigableString, Tag
from lsp_tree_sitter.misc import get_soup

from .._metainfo import SOURCE, project

ALIAS_PREFIX = "(alias: "
PAT = re.compile(r"[a-z-]+")


def get_schema() -> dict[str, Any]:
    r"""Get schema.

    :rtype: dict[str, Any]
    """
    filetype = "tmux"
    schema = {
        "$id": (
            f"{SOURCE}/blob/main/"
            f"src/termux_language_server/assets/json/{filetype}.json"
        ),
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$comment": (
            "Don't edit this file directly! It is generated by "
            f"`{project} --generate-schema={filetype}`."
        ),
        "type": "object",
        "properties": {},
    }
    soup = get_soup("tmux.1", "groff", "mdoc")
    p = soup.find("p", string="CLIENTS AND SESSIONS")
    isoption = 0
    while p and p.text != "EXIT MESSAGES":
        b = p.find("b")
        if (
            isinstance(b, Tag)
            and b.text == p.text.split()[0]
            and not isinstance(p, NavigableString)
            and p.attrs.get("style") == "margin-top: 1em"
        ):
            description = p.text.replace("\n", " ")
            name = b.text
            p = p.find_next("p")
            if not p:
                continue
            b = p.find("b")
            alias = None
            if p.text.startswith(ALIAS_PREFIX) and isinstance(b, Tag):
                alias = b.text
                description += "\n" + description.replace(name, alias)
                p = p.find_next("p")
            if not p:
                continue
            _type = "string"
            if name.endswith("[]"):
                name = name.rstrip("[]")
                _type = "array"
            if not PAT.fullmatch(name):
                continue
            if name == "backspace":
                isoption = 1
                schema["properties"]["set-option"]["properties"] = {}
            s = ""
            enum = []
            if isoption:
                s = "".join(description.split()[1:])
                if s.startswith("[") and s.endswith("]"):
                    enum = [t.strip() for t in s.strip("[]").split("|")]
                description = f"""```tmux
set {description}
```
"""
            else:
                description = f"""```tmux
{description}
```
"""
            description += "\n" + p.text.replace("\n", " ")
            description = description.replace("\u2212", "-")
            if isoption:
                schema["properties"]["set-option"]["properties"][name] = {
                    "description": description,
                    "type": _type,
                }
                if len(enum) > 1:
                    schema["properties"]["set-option"]["properties"][name][
                        "enum"
                    ] = enum
                if s in {
                    "number",
                    "height",
                    "width",
                    "index",
                    "time",
                    "lines",
                }:
                    schema["properties"]["set-option"]["properties"][name][
                        "pattern"
                    ] = "\\d+"
            else:
                schema["properties"][name] = {"description": description}
                if alias:
                    schema["properties"][alias] = {"description": description}
            if name == "window-style":
                isoption = 0
        p = p.find_next("p")
    data = {"type": "array", "uniqueItems": True, "items": {"type": "string"}}
    schema["properties"]["source-file"] |= data
    data = {"patternProperties": {"@[-_\\da-zA-Z]": {"type": "string"}}}
    schema["properties"]["set-option"] |= data
    return schema



r"""Test utils."""

from tmux_language_server.utils import get_schema


class Test:
    r"""Test."""

    @staticmethod
    def test_get_schema() -> None:
        assert get_schema()["properties"].get("set")



r"""Schema
==========
"""

from dataclasses import dataclass

from lsp_tree_sitter import UNI
from lsp_tree_sitter.schema import Trie
from lsprotocol.types import Position, Range
from tree_sitter import Node

DIRECTIVES = {
    "set_option_directive",
    "source_file_directive",
}


@dataclass
class TmuxTrie(Trie):
    r"""Tmux Trie."""

    value: dict[str, "Trie"] | list["Trie"] | str  # type: ignore

    @classmethod
    def from_node(cls, node: Node, parent: "Trie | None") -> "Trie":
        r"""From node.

        :param node:
        :type node: Node
        :param parent:
        :type parent: Trie | None
        :rtype: "Trie"
        """
        if node.type == "value":
            return cls(
                UNI.node2range(node), parent, UNI.node2text(node).strip("'\"")
            )
        if node.type == "file":
            trie = cls(Range(Position(0, 0), Position(1, 0)), parent, {})
            for child in node.children:
                if child.type not in DIRECTIVES:
                    continue
                # directive name
                _type = child.type.split("_directive")[0].replace("_", "-")
                # add directive name to trie.value if it doesn't exist
                _value: dict[str, Trie] = trie.value  # type: ignore
                if _type not in _value:
                    trie.value[_type] = cls(  # type: ignore
                        UNI.node2range(child),
                        trie,
                        {} if _type != "source-file" else [],
                    )
                # the dictionary's key corresponding to directive name
                subtrie: Trie = trie.value[_type]  # type: ignore
                # currently, only support set and source
                # set is a dict, source is a list
                value = subtrie.value  # type: ignore
                # fill subtrie.value
                if child.type == "set_option_directive":
                    value: dict[str, Trie]
                    value[UNI.node2text(child.children[-2])] = cls.from_node(
                        child.children[-1], subtrie
                    )
                elif child.type == "source_file_directive":
                    value += [  # type: ignore
                        cls(
                            UNI.node2range(child.children[1]),
                            subtrie,
                            UNI.node2text(child.children[1]),
                        )
                    ]
            return trie
        raise NotImplementedError(node.type)
